#include "retrohack.h"
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdarg.h>

#ifndef M_PI
#define M_PI 3.141592654f
#endif

// FONT CODE: quick hack!
void *Characters[256];
int CWidths[256];

void InitFonts()
{
	int c = 256;
	while(c--)
		Characters[c] = NULL;
}

void FreeFont()
{
	int c = 256;
	while(c--)
	{
		rt_FreeGraphic(Characters[c]);
		Characters[c] = NULL;
	}
}

int LoadFont(char *name)
{
	FILE *data;
	FreeFont();
	if(!(data = fopen(name, "rb")))
		return FALSE;
		
	while(1)
	{
		int charnum, width, height;
		unsigned char *GData;

		charnum = fgetc(data);
		if(feof(data)) break;

		width = fgetc(data);
		height = fgetc(data);
		GData = (unsigned char *)malloc(sizeof(unsigned char)*(((width+1) >> 1)*height));
		fread(GData, 1, ((width+1) >> 1)*height, data);

		Characters[charnum] = rt_UploadGraphic(RT_Y, GData, 2, width, height, 0);
		CWidths[charnum] = width;

		free(GData);
	}

	fclose(data);
	return TRUE;
}

void FontPrintF(int x, int y, int col, const char *fmt, ...)
{
	char TString[2048], *str = TString;
	va_list ap;

	va_start(ap, fmt);
//	vsnprintf(TString, 2048, fmt, ap);
	vsprintf(TString, fmt, ap);
	va_end(ap);

	while(*str)
	{
		int off = *str;
		rt_MaskedBlit(Characters[off], col >= 0 ? col : rand() / (RAND_MAX >> 4), x, y, itofix(1), itofix(1), 0, 0);
		x += itofix(CWidths[off]);
		str++;
	}
}

int StrWidth(const char *fmt, ...)
{
	char TString[2048], *str = TString;
	int r = 0;
	va_list ap;

	va_start(ap, fmt);
//	vsnprintf(TString, 2048, fmt, ap);
	vsprintf(TString, fmt, ap);
	va_end(ap);

	while(*str)
	{
		int off = *str;
		r += itofix(CWidths[off]);
		str++;
	}
	return r;
}

// GAME CODE: nasty use of globals

unsigned char Board[10][18];
void *BlockGraphics[16];
float Pieces[8][4][2] =
{
	{	{-1.5, -0.5}, {-0.5, -0.5}, {0.5, -0.5}, {1.5, -0.5}	},
	{	{-1, -1}, {-1, 0}, {0, 0}, {1, 0}	},
	{	{1, -1}, {1, 0}, {0, 0}, {-1, 0}	},
	{	{-0.5, -0.5}, {-0.5, 0.5}, {0.5, 0.5}, {0.5, -0.5}	},
	{	{-1, 0}, {0, 0}, {1, 0}, {0, -1}	},
	{	{-1, -1}, {0, -1}, {0, 0}, {1, 0}	},
	{	{1, -1}, {0, -1}, {0, 0}, {-1, 0}	}
};
float CurrentShape[4][2];
Uint8 Colours[] = {1, 3, 5, 7, 4, 13, 15};
float Offsets[] = {0, -0.5, -0.5, 0, 0, -0.5, -0.5};
float CurrentX, CurrentY;

int DropTimes[] = {53, 49, 45, 41, 37, 33, 28, 22, 17, 10, 9, 8, 7, 6, 6, 5, 5, 4, 4, 3};
int CurrentColour, NumLines = 0, Score = 0, DropSpeed, Level = 0;

void ClearBoard(void)
{
	int x, y;
	for(x = 0; x < 10; x++)
		for(y = 0; y < 18; y++)
			Board[x][y] = 0;
}

void SetColour(float x, float y, int colour)
{
	int ix = (int)floor(x), iy = (int)floor(y);
	if(ix >= 0 && ix < 10 && iy >= 0 && iy < 18)
		Board[(int)floor(x)][(int)floor(y)] = colour;
}

int GetColour(float x, float y)
{
	int ix = (int)floor(x), iy = (int)floor(y);
	if(ix >= 0 && ix < 10 && iy >= 0 && iy < 18)
		return Board[ix][iy];

	return ((iy < 0) && (ix >= 0) && (ix < 10)) ? 0 : -1;
}

void DrawPiece(int colour)
{
	int c = 4;
	while(c--)
		SetColour(CurrentX + CurrentShape[c][0], CurrentY + CurrentShape[c][1], colour);
}

int Overlapping(void)
{
	int c = 4;
	while(c--)
	{
		if(GetColour(CurrentX + CurrentShape[c][0], CurrentY + CurrentShape[c][1]))
			return TRUE;
	}
	return FALSE;
}

void PickPiece(void)
{
	int piece = ((rand() >> 4) * 7) / (RAND_MAX >> 4);
	CurrentY = 0;
	CurrentX = 5 + Offsets[piece];
	memcpy(CurrentShape, &Pieces[piece], 4*2*sizeof(float));
	CurrentColour = Colours[piece];
}

int NewPiece(void)
{
	PickPiece();
	return Overlapping();
}

void RotateShape(float xmul, float ymul)
{
	int c = 4;
	while(c--)
	{
		float NewY = ymul * CurrentShape[c][0];
		CurrentShape[c][0] = xmul * CurrentShape[c][1];
		CurrentShape[c][1] = NewY;
	}
}

void DrawBoard(int Colour)
{
	int x, y;
	for(x = 0; x < 10; x++)
		for(y = 0; y < 18; y++)
			if(Board[x][y])
				rt_Blit(BlockGraphics[Colour], Board[x][y], itofix(3 + (x * 13)), itofix(3 + (y * 13)), itofix(13), itofix(13), 0, 0);
}

void MoveBlock(float dir)
{
	DrawPiece(0);
	CurrentX += dir;
	if(Overlapping()) CurrentX -= dir;
	DrawPiece(CurrentColour);
}

void Rotate(int direction)
{
	DrawPiece(0);
	RotateShape(direction, -direction);
	if(Overlapping()) 
		RotateShape(-direction, direction);
	DrawPiece(CurrentColour);
}

void AddLines(int NewLines)
{
	int Scores[] = {0, 40, 100, 300, 1200};

	NumLines += NewLines;
	Score += Scores[NewLines]*(Level + 1);
	if(Level < NumLines/10)
		Level = NumLines / 10;
	if(Level > 19) Level = 19;
	DropSpeed = DropTimes[Level];
}

void GrabFont()
{
	FILE *output;
	int bgcolour;
	int scanx, scany, charnum = 32;

	BITMAP *b = load_pcx("tahoma.pcx", NULL);
	blit(b, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
	bgcolour = getpixel(b, 0, 0);
	scanx = scany = 0;
	output = fopen("tahoma.fnt", "wb");
	
	// go left to right, top to bottom
	while(scany < b->h)
	{
		int endy;

		// find new top
		while(1)
		{
			for(scanx = 0; scanx < b->w; scanx++)
			{
				if(getpixel(b, scanx, scany) != bgcolour)
					break;
			}
			if(scanx == b->w)
				scany++;
			else
				break;
		}
		
		if(scany >= b->h)
			break;

		// find new bottom
		endy = scany;
		while(1)
		{
			for(scanx = 0; scanx < b->w; scanx++)
			{
				if(getpixel(b, scanx, endy) != bgcolour)
					break;
			}
			if(scanx == b->w)
				break;
			else
				endy++;
		}

		printf("%d to %d\n", scany, endy);

		// read 16 characters
		scanx = 0;
		while(scanx < b->w)
		{
			int endx, ix, iy;
			int newcol;
//			int bytes;

			// find next start
			while(getpixel(b, scanx, scany) == bgcolour)
				scanx++;
				
			if(scanx >= b->w) break;

			// find next end
			endx = scanx;
			while(getpixel(b, endx, scany) != bgcolour)
				endx++;

			// save to file
			fputc(charnum, output); charnum++;
			fputc(endx - scanx, output);
			fputc(endy - scany, output);
			newcol = 0;

			for(iy = scany; iy < endy; iy++)
			{
				for(ix = scanx; ix < endx; ix++)
				{
					newcol = (newcol << 4) | (getpixel(b, ix, iy) >> 4);
					if((ix - scanx)&1) fputc(newcol&0xff, output); 
				}
				if((ix - scanx)&1) fputc((newcol&0xf) << 4, output); 
			}
			ix = endx - scanx; iy = endy - scany;
			scanx = endx;
		}

		scany = endy;
	}

	fclose(output);
	readkey();
	destroy_bitmap(b);
}

int WaitKeyPressed = FALSE;

void ClearKeyPressed(void)
{
	WaitKeyPressed = TRUE;
}

int KeyPressed(void)
{
	int Res = 0xff;
	int c = 11;
	while(c--)
		Res &= rt_ReadKeyboard(c);

	if(Res == 0xff && WaitKeyPressed)
		WaitKeyPressed = FALSE;

	return Res != 0xff && !WaitKeyPressed;
}

int main(int argc, const char *argv[])
{
	void *RTLogo, *back, *GameLogo = NULL, *BackDrop = NULL;
	unsigned char Data[320 * 240];
	FILE *gfxfile;
	int x, BoardColour = 0;
	int BKColour = 15;
	enum GameStates
	{
		GS_TITLE1, GS_TITLE2, GS_INTRO, GS_GAME, GS_LINECHECKING, GS_LINECLEARING, GS_GAMEOVER
	} CurState = GS_INTRO;
	char ScrollText[] =
		"                       Hello, and welcome to the RetroHack demo game. It's not very good, I'm afraid - just a slightly tedious clone of circa 1990 Tetris. See some of the other demos for some more advanced effects.                       The wonderful soundtrack to this game has been lifted from Tetris 2 by Frantisek Fuka. Everything else is my own, for the little it's worth.                       \
		I guess this is a good place to recap the RetroHack machine and the competition rules.                       The RetroHack machine has an AY-3-8910 sound generator and a custom blitter, based heavily on the Atari Lynx's Suzy.                       \
		The AY-3-8910 was used in several real machines, including the Atari ST, Amstrad CPC, 128k ZX Spectrums and the MSX.                       Suzy was only used in the Atari Lynx, but comes from an identifiable series of machines in which it is preceeded by the Amiga and followed by the 3do.                       Now some competition rules.                       \
		Entries must be submitted by 12 midnight GMT on Sunday the 11th of November. They must consist of source and data which, when unarchived, occupy no more than 512 kB of disk space. No libraries other than the RetroHack library and the standard compiler libraries may be used. Entries must be submitted in either C or C++.                       \
		The RetroHack simulator does two non-obvious things: it switches into and out of fullscreen mode on alt+enter, and automatically frees CPU time so that it doesn't use too much more than it actually needs.                                              \
		That's all. If you plan to compete then good luck!                                              ";
	int ScrollOff = 0, ScrollPtr = 0;
	int InGame = FALSE;

	rt_Init();
	rtSim_SetWindowTitle("RetroHack: Tetrominoes");
	InitFonts();

	// load RetroHack logo
	gfxfile = fopen("logo.raw", "rb");
	fread(Data, 1, 320*42, gfxfile);
	fclose(gfxfile);

	for(x = 0; x < 320 * 42; x++)
		Data[x] >>= 4;

	RTLogo = rt_UploadGraphic(RT_Y, Data, 1, 320, 42, 0xf);
	psg_OpenFile("menu1.psg", FALSE);

	Data[0] = 10;
	back = rt_UploadGraphic(RT_CbCr, Data, 1, 1, 1, 0);

	x = 16;
	while(x--)
	{
		Data[0] = x;
		BlockGraphics[x] = rt_UploadGraphic(RT_Y, Data, 1, 1, 1, 0);
	}

	LoadFont("tahoma.fnt");

	while(!rtSim_QuitWanted())// && (rt_ReadKeyboard(0)&0x02))
	{
		enum GameStates NextState = CurState;
		static int NewLines = 0;
		static int UpdateY;

		if(!InGame)
			rt_Blit(back, BKColour, 0, 0, 320 << 16, 240 << 16, 0, 0);
		else
		{
			rt_Blit(BackDrop, 10, 0, 0, itofix(1), itofix(1), 0, 0);
			DrawBoard(BoardColour >> 4);
			FontPrintF(itofix(180), itofix(3), 9, "Score: %d", Score);
			FontPrintF(itofix(180), itofix(33), 9, "Lines: %d", NumLines);
			FontPrintF(itofix(180), itofix(63), 9, "Level: %d", Level);
		}

		switch(CurState)
		{
			case GS_INTRO:
			{
				static float angle = 0;
				static float angleadd = 0;
				float xw = cos(angle), yw = sin(angle);
				rt_MaskedBlit(RTLogo, 10, itofix(160) - ftofix(xw * 160), itofix(120) - ftofix(yw * 21), ftofix(xw), ftofix(yw), 0, 0);
				angle += angleadd;
				angleadd += 0.00011f;
				if(angle > M_PI)
					NextState = GS_TITLE1;
				if(!(rt_ReadKeyboard(8)&0x80)) NextState = GS_GAME;
			}
			break;
			case GS_GAMEOVER:
				FontPrintF(itofix(160) - (StrWidth("GAME OVER") >> 1), itofix(120 - 13), 10, "GAME OVER");
				BoardColour -= 3;
				if(BoardColour < 16) BoardColour = 16;

				// if new key pressed, return to main menu
				if(KeyPressed())
				{
					InGame = FALSE;
					NextState = GS_TITLE2;
				}
			break;
			case GS_TITLE1:
			{
				static int Time = 0;

				if(Time < 15)
				{
					BKColour--;
				}
				else
				{
					int TimeOffs;
					if(Time < 160)
					{
						TimeOffs = (160 - Time) *100;
					}
					else
						TimeOffs = 0;
					rt_MaskedBlit(GameLogo, 10, ftofix(62.5f), itofix(10), itofix(1), itofix(1) + TimeOffs*4, -TimeOffs * 63, TimeOffs);
				}

				Time++;
				if(!(rt_ReadKeyboard(8)&0x80)) NextState = GS_GAME;
			}
			break;
			case GS_TITLE2:
			{
				char ST[2048];
				int copyc = 0;

				// show logo
				rt_MaskedBlit(GameLogo, 10, ftofix(62.5f), itofix(10), itofix(1), itofix(1), 0, 0);

				// show instructions
				FontPrintF(itofix(30), itofix(60), 10, "LEFT/RIGHT: move block");
				FontPrintF(itofix(30), itofix(86), 10, "DOWN: fast drop");
				FontPrintF(itofix(30), itofix(112), 10, "UP: rotate (anti-clockwise)");
				FontPrintF(itofix(30), itofix(138), 10, "ESC: return to title screen");
				FontPrintF(itofix(30), itofix(190), 10, "SPACE: start game");

				// show scroller
				ScrollOff -= itofix(5);
				if(!ScrollText[ScrollPtr]) ScrollPtr = 0;
				while(1)
				{
					ST[copyc] = ScrollText[copyc+ScrollPtr];
					copyc++;
					ST[copyc] = '\0';
					if(StrWidth(ST) + ScrollOff < 0)
					{
						ScrollPtr++;
						ScrollOff += StrWidth(ST);
						copyc = 0;
					}
					if(StrWidth(ST) + ScrollOff > itofix(320) || !ScrollText[copyc+ScrollPtr]) break;
				}
				FontPrintF(ScrollOff, itofix(214), -1, ST);

				if(KeyPressed() && !(rt_ReadKeyboard(8)&0x80)) NextState = GS_GAME;
			}
			break;
			case GS_GAME:
			{
				static int Pause = 25;
				static int PauseP = 0;
				static Uint8 OldKeyb10 = 0;
				Uint8 NewKeyb10, Differences;

				// check keys
				if(!(rt_ReadKeyboard(0)&0x02))	// ESC: back to title screen
				{
					InGame = FALSE;
					NextState = GS_TITLE2;
				}

				NewKeyb10 = rt_ReadKeyboard(10)^0xff;
				Differences = NewKeyb10 ^ OldKeyb10;

				if(NewKeyb10&Differences&0x10) MoveBlock(-1);	// LEFT
				if(NewKeyb10&Differences&0x80) MoveBlock(1);	// RIGHT
				if(NewKeyb10&Differences&0x40) Rotate(1); // UP
				if(NewKeyb10&0x20 && Pause > (1 + PauseP)) Pause = 1 + PauseP;	// DOWN

				OldKeyb10 = NewKeyb10;

				AddLines(NewLines); NewLines = 0;
				Pause--;
				if(!Pause)
				{
					PauseP ^= 1;
					DrawPiece(0);
					CurrentY++;
					if(Overlapping())
					{
						CurrentY--;
						DrawPiece(CurrentColour);
						UpdateY = 17;
						NextState = GS_LINECHECKING;
						Pause = 1;
					}
					else
					{
						DrawPiece(CurrentColour);
						Pause = DropSpeed;
					}
				}
			}
			break;
			case GS_LINECLEARING:
			{
				static int Pause = 3;
				Pause--;
				if(!Pause)
				{
					int x, y;
					for(x = 0; x < 10; x++)
					{
						for(y = UpdateY; y > 0; y--)
							SetColour(x, y, GetColour(x, y-1));
						SetColour(x, 0, 0);
					}

					Pause = 3;
					NextState = GS_LINECHECKING;
				}
			}
			break;
			case GS_LINECHECKING:
			{
				while(UpdateY >= 0)
				{
					int c = 0, x;
					for(x = 0; x < 10; x ++)
					{
						if(GetColour(x, UpdateY))
							c++;
					}
					if(c == 10)
						break;
					UpdateY--;
				}

				if(UpdateY >= 0)
				{
					for(x = 0; x < 10; x ++)
						SetColour(x, UpdateY, 10);
					NextState = GS_LINECLEARING;
					NewLines++;
				}
				else
				{
					NextState = NewPiece() ? GS_GAMEOVER : GS_GAME;
				}
			}
			break;
		}

		// update sound
		rt_Flip();
		if(!psg_Update())
			NextState = GS_TITLE2;

		// prepare new state if necessary
		if(NextState != CurState)
		{
			if(RTLogo) {rt_FreeGraphic(RTLogo); RTLogo = NULL;}
			if(!GameLogo)
			{
				// load (and invert) game logo (195 x 26)
				gfxfile = fopen("Tetrominoes.raw", "rb");
				fread(Data, 1, 195*26, gfxfile);
				fclose(gfxfile);
				for(x = 0; x < 195 * 26; x++)
					Data[x] = (Data[x] >> 4)^0xf;
				GameLogo = rt_UploadGraphic(RT_Y, Data, 1, 195, 26, 0);
				
				gfxfile = fopen("beach.raw", "rb");
				fread(Data, 1, 320*240, gfxfile);
				fclose(gfxfile);
				for(x = 0; x < 320 * 240; x++)
					Data[x] >>= 4;
				BackDrop = rt_UploadGraphic(RT_Y, Data, 1, 320, 240, 0);
			}

			switch(NextState)
			{
				default: break;
				case GS_GAMEOVER:
					ClearKeyPressed();
				break;
				case GS_TITLE2:
				{
					ScrollPtr = 0;
					psg_OpenFile("menu2.psg", TRUE);
					BKColour = 0;
					ScrollOff = 0;
					ClearKeyPressed();
				}
				break;
				case GS_GAME:
					InGame = TRUE;
					if(CurState != GS_LINECHECKING)
					{
						BoardColour = 255;
						NumLines = 0;
						Score = 0;
						DropSpeed = DropTimes[0];
						Level = 0;
						BKColour = 0;
						psg_OpenFile("game.psg", TRUE);
						ClearBoard();
						PickPiece();
					}
				break;
			}
			CurState = NextState;
		}
	}

	psg_CloseFile();
	FreeFont();
	rt_FreeGraphic(back);
	x = 16;
	while(x--)
		rt_FreeGraphic(BlockGraphics[x]);
	if(GameLogo) rt_FreeGraphic(GameLogo);
	if(BackDrop) rt_FreeGraphic(BackDrop);
	rt_Exit();

	return 0;
}
END_OF_MAIN()
